/** @file ddcutil_macros.h.in
 *
 *  Macros for checking ddcutil version at compile time. 
 *  The Autotools build system converts this file to ddcutil.macros.h. 
 *  The version values are set from those in configure.ac. 
 */

// Copyright (C) 2018-2021 Sanford Rockowitz <rockowitz@minsoft.com>
// SPDX-License-Identifier: GPL-2.0-or-later


#define DDCUTIL_VMAJOR  1
#define DDCUTIL_VMINOR  5
#define DDCUTIL_VMICRO  0
#define DDCUTIL_VSUFFIX dev

